from base64 import b64encode

import json
from urllib.request import urlopen, Request

from .error import HTTPError

error_msgs = {
    400: "parameter error",
    401: "authorization error",
    404: "api not found",
    500: "internal error",
    503: "system resources is not sufficient. It is may be caused by a huge query."
}


class Client:
    def __init__(self, url: str = "http://localhost:31007/",
                 database: str = "public",
                 user: str = "root",
                 password: str = ""):
        self._url = url
        self._database = database
        self._user = user
        self._password = password

    def sql(self, q: str) -> dict:
        _url = self._url + f"api/v1/sql?db={self._database}&pretty=true"
        request = self.build_request(url=_url, data=q)
        response = urlopen(request)
        self._check_status(response)
        resp = response.read().decode('utf-8')
        if resp == "":
            return {}
        resp = json.loads(resp)
        return resp

    def line_protocol(self, lines: [str]):
        _url = self._url + f"api/v1/write?db={self._database}&pretty=true"
        _data = ""
        for line in lines:
            _data += line + "\n"
        request = self.build_request(url=_url, data=_data)
        response = urlopen(request)
        self._check_status(response)

    def set_database(self, database_name):
        self._database = database_name

    def set_user(self, user, password):
        self._user = user
        self._password = password

    def set_url(self, url):
        self._url = url

    def _check_status(self, response):
        status = response.status
        if status != 200:
            msg = error_msgs.get(status)
            raise HTTPError(status, msg)

    def build_request(self, url, data):
        _basic = basic_auth(self._user, self._password)
        request = Request(url=url, data=data.encode("UTF-8"))
        request.add_header("Authorization", "Basic " + _basic)
        request.add_header("Accept", "application/json")
        return request


def basic_auth(name, pwd):
    return b64encode((name + ":" + pwd).encode(encoding="UTF-8")).decode(encoding="UTF-8")



from .error import *
from .client import Client


# PEP249 cursor
class Cursor:
    def __init__(self, client: Client):
        self._c = client
        self.arraysize = 1
        self._rowcount = -1
        self._response = None
        self._index = -1
        self._description = None

    @property
    def rowcount(self):
        return self._rowcount

    @property
    def description(self):
        """
        :return: each item is a list contain two str item
                which is column name and column type
        """
        if self._response is None:
            return None
        return self._description

    def callproc(self, procname, parameters=None):
        raise NotSupportedError()

    def close(self):
        pass

    def execute(self, operation, parameters=None):
        self._response = None
        self._index = -1
        self._response = self._c.sql(operation)
        self._rowcount = len(self._response)
        description = []
        for row in self._response:
            for column in row:
                description.append([column, type(row[column]).__name__])
            break
        self._description = description

    def executemany(self, operation, parameters=None):
        self.execute(operation)

    def get_type(self, col):
        if not self._description:
            return None
        return self._description[col][1]

    def fetchone(self):
        if self._response is None:
            raise OperationalError("no result to fetch")
        self._index += 1
        if self._index + 1 > len(self._response):
            return None
        return self._response[self._index]

    def fetchmany(self):
        return self.fetchone()

    def fetchall(self):
        if self._response is None:
            raise OperationalError("no result to fetch")
        start_index = self._index + 1
        self._index = len(self._response)
        return self._response[start_index:]

    def nextset(self):
        raise NotSupportedError()

    def setinputsizes(self):
        raise NotSupportedError()

    def setoutputsize(self, size, column=None):
        raise NotSupportedError()



from .conn import CnosDBConnection


def connect(**kwargs) -> CnosDBConnection:
    """
    Keyword Arguments
    ========================================================================
    url: str
        optional, default_value: "http://127.0.0.1:31007/"
        description: the url to connect db server

    database: str
        optional, default_value: public
        description: database name for client to send message

    user: str
        optional, default_value: "root"
        description: default user

    password: str
        optional, default_value: ""
        description: the password of user, default user root password is ""
    ========================================================================
    ```
    example:
        import cnos_connector

        conn = cnos_connector.connect(url="http://192.168.0.11/", database="public", user="root", password="123456")
    ```
    """
    return CnosDBConnection(**kwargs)



from .error import NotSupportedError
from .client import Client
from .cursor import Cursor


class CnosDBConnection:
    def __init__(self, **kwargs):
        url = kwargs.get("url", "http://127.0.0.1:31007/")
        user = kwargs.get("user", "root")
        password = kwargs.get("password", "")
        database = kwargs.get("database", "public")
        self._client = Client(
            url,
            database,
            user,
            password,
        )

    # close, commit, rollback, cursor are PEP249 method
    def close(self):
        pass

    def commit(self):
        pass

    def rollback(self):
        raise NotSupportedError()

    def cursor(self):
        return Cursor(self._client)

    def execute(self, sql: str):
        """
        database execute sql

        :param sql: the CnosDB sql, more information for CnosDB can be found in
                    https://docs.cnosdb.com/guide/introduction.html
        :type sql: str
        :return: the CnosDB execute sql result
        :rtype: dict list, the sql result of CnosDB
        """
        return self._client.sql(sql)

    def write_lines(self, lines: [str]):
        """
        write lineprotocol to database

        :param lines: The lines of LineProtocol
                      each item is lineprotocol str, more information for LineProtocol can be found in
                      https://docs.influxdata.com/influxdb/v2.6/reference/syntax/line-protocol/
        """
        self._client.line_protocol(lines)

    def write_csv(self, table, file_path):
        """
        write a csv file to database

        :param table: the table for csv file to write
                      before write csv file to database, database should
                      have a table which have same schema with csv
        :type table: str
        :param file_path: file path of the csv file
        :type file_path: str
        """
        # create a temp external table use csv file, then get table
        # schema of given table, through INSERT SELECT write data
        # to given table
        self._client.sql(f"create external table ___temp_external_table\
                            STORED AS CSV \
                            WITH HEADER ROW \
                            LOCATION \"{file_path}\";")
        schema = self._client.sql(f"DESCRIBE TABLE {table}")
        col_names = ""
        for column in schema:
            col_names += column["COLUMN_NAME"] + ","
        col_names = col_names.strip(',')
        self._client.sql(f"INSERT INTO {table}({col_names}) SELECT * FROM ___temp_external_table;")
        self._client.sql("DROP TABLE ___temp_external_table")

    def create_database(self, database_name):
        """
        create a database from CnosDB

        :param database_name: the name of database to create
        :type database_name: str
        """
        self._client.sql(f"CREATE DATABASE {database_name};")

    def create_database_with_ttl(self, database_name, ttl):
        """
        create a database with ttl from CnosDB

        :param database_name: the name of database to create
        :type database_name: str
        :param ttl: time to live, means the data time to live in database,
                    default is 365 day, when set ttl to 30 day, then write
                    data with timestamp for before 30 day will be failed
        :type ttl: duration str, use number and unit, support d(day), h(hour), m(minutes),
                    such as "40d", "100h", "3600m"
        :return:
        """
        self._client.sql(f"CREATE DATABASE {database_name} WITH TTL '{ttl}';")

    def create_user(self, user, password):
        """
        create a user from CnosDB

        :param user: the username of database to create
        :type user: str
        :param password: the password of the user
        :type password: str
        :return:
        """
        self._client.sql(f"CREATE USER {user} WITH PASSWORD = {password}")

    def drop_database(self, database_name):
        """
        drop a database from CnosDB

        :param database_name: the database name of CnosDB to drop
        :type database_name: str
        :return:
        """
        self._client.sql(f"DROP DATABASE {database_name};")

    def drop_table(self, table_name):
        """
        drop a table from CnosDB

        :param table_name: the table name of database to drop
        :type table_name: str
        :return:
        """
        self._client.sql(f"DROP TABLE {table_name};")

    def drop_user(self, user):
        """
        drop a user from CnosDB

        :param user: the username of database to drop
        :type user: str
        :return:
        """
        self._client.sql(f"DROP USER {user}")

    def switch_database(self, database_name):
        """
        switch client current database to given database

        :param database_name: the database name of client to switch
        :type database_name: str
        :return:
        """
        self._client.set_database(database_name)

    def switch_user(self, user, password):
        """
        switch client current database to given user

        :param user: the username of client to switch
        :type user: str
        :param password: the password of the user
        :type password: str
        :return:
        """
        self._client.set_user(user, password)

    def switch_url(self, url):
        """
        switch client current url to given url

        :param url: the url of client to switch
        :type url: str
        :return:
        """
        self._client.set_url(url)

    def list_database(self):
        """
        list all database in CnosDB

        :return: the database names in CnosDB
        :rtype: dict list
        """
        self._client.sql("SHOW DATABASES")

    def list_table(self):
        """
        list all table in client current database

        :return: the table names in database
        :rtype: dict list
        """
        self._client.sql("SHOW TABLES")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cnosdb_connector']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'cnos-connector',
    'version': '0.1.4',
    'description': 'CnosDB Python Client',
    'long_description': '# CnosDB Python Connector\n\nCnosDB Python Connector repository contains the Python client library for the [CnosDB](https://github.com/cnosdb/cnosdb). cnosdb_connector adapted for PEP249.\n\n## Installation\n\nuse pip install it from pypi, **Python 3.6** or later is required.\n\n```\npip install cnos-connector\n```\n\nThen import the package:\n\n```\nimport cnosdb_connector\n```\n\n## Getting Started\n\n#### Query use SQL\n\n```python\nfrom cnosdb_connector import connect\n\nconn = connect(url="http://127.0.0.1:31007/", user="root", password="")\nresp = conn.execute("SHOW DATABASES")\nprint(resp)\n```\n\n#### Query use interface\n\n```python\nfrom cnosdb_connector import connect\n\nconn = connect(url="http://127.0.0.1:31007/", user="root", password="")\nconn.create_database("air")\nresp = conn.list_database()\nprint(resp)\n```\n\n#### Query use PEP-249\n\n```python\nfrom cnosdb_connector import connect\n\nconn = connect(url="http://127.0.0.1:31007/", user="root", password="")\ncursor = conn.cursor()\n\ncursor.execute("SHOW DATABASES")\nresp = cursor.fetchall()\nprint(resp)\n```\n\n#### Query use pandas\n\n```python\nimport pandas as pd\nfrom cnosdb_connector import connect\n\nconn = connect(url="http://127.0.0.1:31007/", user="root", password="")\n\nresp = pd.read_sql("SHOW DATABASES", conn)\nprint(resp)\n```\n\n#### Write use LineProtocol\n\n```python\nfrom cnosdb_connector import connect\n\nline0 = "test_insert,ta=a1,tb=b1 fa=1,fb=2 1"\nline1 = "test_insert,ta=a1,tb=b1 fa=3,fb=4 2"\nline2 = "test_insert,ta=a1,tb=b1 fa=5,fb=6 3"\n\nconn = connect(url="http://127.0.0.1:31007/", user="root", password="")\n\nconn.create_database_with_ttl("test_database", "100000d")\nconn.switch_database("test_database")\n\nconn.write_lines([line0, line1, line2])\n\nresp = conn.execute("SELECT * FROM test_insert;")\nprint(resp)\n```\n\n#### Write use SQL\n\n```python\nfrom cnosdb_connector import connect\n\nconn = connect(url="http://127.0.0.1:31007/", user="root", password="")\n\nquery = "insert test_insert(TIME, column1, column2, column3, column4, column5, column6, column7) values (100, -1234, \'hello\', 1234, false, 1.2, \'beijing\', \'shanghai\'); "\n\nconn.execute(query)\n\nresp = conn.execute("SELECT * FROM test_insert;")\nprint(resp)\n```\n\n#### Write use CSV\n\n```python\nfrom cnosdb_connector import connect\nimport os\n\nquery = "CREATE TABLE test_insert(column1 BIGINT CODEC(DELTA),\\\n                                  column2 BOOLEAN,\\\n                                  column3 DOUBLE CODEC(GORILLA),\\\n                                  TAGS(column4));"\n\nconn = connect(url="http://127.0.0.1:31007/", user="root", password="")\n# table schema must same with csv file\nconn.execute(query)\n\npath = os.path.abspath("test.csv")\nconn.write_csv("test_insert", path)\n\nresp = conn.execute("SELECT * FROM test_insert;")\nprint(resp)\n```\n\n# License\n\nCnosDB Python Connector use MIT License',
    'author': 'Subsegment',
    'author_email': '304741833@qq.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
